
{- A sample run


chameleon -d --no-term --no-rr samples/collects.ch
_______________
-- Chameleon --

  Checking locations [239,268,278,282,291,297,320]
  Well-typed
  Solver Mode: 0
  Justification Policy: local
collects.ch> type l
320) l  = [insert empty A, e2, insert e1 B, empty, c, f A] :: Insert(a -> T -> a), Empty(a) => [a]


Options --no-term and --no-rr are briefly explained below.

-}


---- Encoding of multi-parameter class:
--
-- class Collects ce e | ce -> e where
--     empty  :: ce
--     insert :: ce -> e -> ce
--
-- instance Collects [e] e where
--     empty       = []
--     insert es e = e:es


---- Constraint abbreviations are not implemented yet...
--
-- constraint Collects (ce,e) = Empty ce, Insert (ce->e->ce);
--
-- rule Empty x  ==> x = ce, Collects (ce,e);
-- rule Insert x ==> x = ce->e->ce, Collects (ce,e);
-- rule Collects (ce,e), Collects (ce,e') ==> e = e';
--
---- Thus:


rule Empty x  ==> x = ce, Empty ce, Insert (ce->e->ce)
rule Insert x ==> x = ce->e->ce, Empty ce, Insert (ce->e->ce)

{- The above rules are only necessary to ensure that all methods
   of a class are indeed present. Because of these rules we need to
   call Chameleon with the --no-term option.

    --no-term  don't check for termination

    Although the CHRs are non-termination our termination checker suggests
    that the CHRs are non-terminating. Obviously, any decidable termination
    checker must be incomplete.

    --no-rr   don't check for range-restrictedness

    A CHR is range-restricted iff all variables on the rhs appear also on the lhs.
    For example, 

      rule Empty x  ==> x = ce, Empty ce, Insert (ce->e->ce);

    is NOT range-restricted. `ce` appears on the rhs but NOT on the lhs.
-}


rule Empty ce, Insert (ce->e->ce), Empty ce, Insert (ce->e'->ce) ==> e = e'


--rule Empty ce, Insert (ce->e->ce') ==> ce'=ce
rule Insert x ==> x=ce->e->ce


rule Insert (ce->e->ce), Insert (ce->e'->ce) ==> e = e'
rule Insert ([e]->e'->[e]) ==> e'=e
-- the two rules above are neccessary to complete the CHRs
-- we're working on automatic completion of CHRs


overload empty :: [e] where
         empty = []
overload insert :: [e] -> e -> [e] where
         insert es e = e:es


-- Test:

data T = A | B

e1 = empty
e2 = empty  
c  = insert e1 A
f  = insert e2
l  = [insert empty A, e2, insert e1 B, empty, c, f A]
