
{-
   chameleon -d --no-term samples/enum.ch

 Option --no-term is briefly explained below

-}

---- Encoding of Enum class:
--
-- class Enum a where
--     pred, succ :: a -> a
--     toEnum :: Int -> a
--
-- instance Enum Int where
--     pred n = n - 1
--     succ n = n + 1
--     toEnum n = n

---- Simulate integers

data Nat = Zero | Succ Nat
data Int = Pos Nat | Neg Nat  -- Neg n with n <> Zero

predInt (Pos (Succ n))		= Pos n
predInt (Pos Zero)		= Neg (Succ Zero)
predInt (Neg n)			= Neg (Succ n)

succInt (Pos n)			= Pos (Succ n)
succInt (Neg (Succ Zero))	= Pos Zero
succInt (Neg (Succ n))		= Neg n

---- Constraint abbreviations are not implemented yet...
--
-- constraint Enum a = Pred (a->a), Succ (a->a), ToEnum (Int->a);
--
-- rule Pred x   ==> x = a->a, Enum a
-- rule Succ x   ==> x = a->a, Enum a
-- rule ToEnum x ==> x = Int->a, Enum a
--
---- Thus:


rule Pred x   ==> x = a->a,   Pred (a->a), Succ (a->a), ToEnum (Int->a)
rule Succ x   ==> x = a->a,   Pred (a->a), Succ (a->a), ToEnum (Int->a)
rule ToEnum x ==> x = Int->a, Pred (a->a), Succ (a->a), ToEnum (Int->a)

{- The above rules are only necessary to ensure that all methods
   of a class are indeed present. Because of these rules we need to
   call Chameleon with the --no-term option.

    --no-term  don't check for termination

    Although the CHRs are non-termination our termination checker suggests
    that the CHRs are non-terminating. Obviously, any decidable termination
    checker must be incomplete.

-}


overload pred :: Int -> Int where
         pred = predInt
overload succ :: Int -> Int where
	 succ = succInt
overload toEnum :: Int -> Int where
         toEnum n = n


-- Example:

map f []	= []
map f (x:xs)	= f x : map f xs

f = map succ
