
---- Encoding of Functor class:
--
-- class Functor f where
--     fmap :: (a -> b) -> (f a -> f b)
--
-- instance Functor [] where
--     fmap f []     = []
--     fmap f (x:xs) = f x : fmap f xs


-- Modelling kind checking in Chamleon

-- Predicate kind
overload kind

data Star = Star

-- We introduce internal encodings of
data L a = L a   -- list,
data A a = A a   -- arrow, and
data P a = P a   -- pair constructor

-- We need them, because in our current syntax we don't allow for partial application of
-- arrow and pair constructors. Therefore, we will transform types from external to
-- internal syntax. See below.

-- Kinding rules

rule Kind (A,k) ==> k=Star->Star->Star
rule Kind (P,k) ==> k=Star->Star->Star
rule Kind (L,k) ==> k=Star->Star

-- Booleans are builtin
rule Kind (Bool,k) ==> k=Star

-- Handling constructor application
rule Kind (c c',k'') ==> Kind (c,k), Kind (c',k'), k=k'->k''

-- Transforming external into internal syntax for kind checking
overload trans;

rule Trans (t->t') <==> Kind ((A t) t',k)
rule Trans ((t,t')) <==> Kind ((P t) t',k)
rule Trans ([t]) <==> Kind (L t,k)

-- For each user-defined constructor we need to 

data Int = One
-- provide the appropriate kinding rule and
rule Kind (Int,k) ==> k=Star

-- Two transformation rules, constructor might appear in argument or function context
rule Trans (Int) <==> Kind (Int,k)
rule Trans (Int t) <==> Kind (Int t,k)


overload fmap

-- We apply this rule for kind checking
--rule Fmap a ==> a=(b->c)->(f b->f c), Trans ((b->c)->(f b->f c))

-- This rule mimics Haskell-style functor classes
rule Fmap a ==> a=(b->c)->(f b->f c)

overload fmap :: (a -> b) -> [a] -> [b] where
         fmap f []     = []
	 fmap f (x:xs) = f x : fmap f xs


-- Example:

data A = A1 | A2
data B = B1 | B2

f A1 = B2
f A2 = B1

x = fmap f [A1,A1,A2,A1,A2]
g = fmap f

id x = x

overload one

overload num

rule One t <==> Num t

test = fmap id one
