
-- helpers
data Int = Z | S Int
fst (a,b) = a
snd (a,b) = b

-- language of object expressions
data ExpVar x      = ExpVar x
data ExpConst n    = ExpConst n
data ExpPair a1 a2 = ExpPair a1 a2
data ExpPi1 a      = ExpPi1 a
data ExpPi2 a      = ExpPi2 a
data ExpAbs x a    = ExpAbs x a
data ExpApp a1 a2  = ExpApp a1 a2
data ExpLet x e1 e2 = ExpLet x e1 e2
data ExpPLet x e1 e2 = ExpPLet x e1 e2
data ExpLVar x       = ExpLVar x
data D a = D a

-- result of type-level computations
data T = T
data F = F

-- we use numbers to model variables
data Zero   = Zero
data Succ n = Succ n

overload eq :: Zero -> Zero -> T
   where eq Zero Zero = T
overload eq :: (Eq (a->b->c)) => Succ a -> Succ b -> c
   where eq (Succ a) (Succ b) = eq a b
overload eq :: Zero -> Succ a -> F
   where eq Zero (Succ a) = F
overload eq :: Succ a -> Zero -> F
   where eq (Succ a) Zero = F

{- these rules are important otherwise we can't check
   the inferred type of lookup against the declared type -}
rule Eq (a->b->c), Eq (a->b->d) ==> c=d
rule Eq (Zero->Zero->a)         ==> a=T
rule Eq (Succ a->Zero->b)       ==> b=F 
rule Eq (Zero->Succ a->b)       ==> b=F     


-- we use lists for the environment
data Nil      = Nil
data Cons a b = Cons a b


overload lookup :: (Lookup' (Cons (x1,v) e -> x2 -> b -> v'), Eq (x1->x2->b))
                => Cons (x1,v) e -> x2 -> v'
   where lookup (Cons (x1,v) e) x2 = lookup' (Cons (x1,v) e) x2 (eq x1 x2)
overload lookup' :: Cons (x1,v) e -> x2 -> T -> v
   where lookup' (Cons (_,v) e) _ T = v
overload lookup' :: (Lookup (e->x2->v')) => Cons (x1,v) e -> x2 -> F -> v'
   where lookup' (Cons (x1,v) e) x2 F = lookup e x2

rule Lookup' (Cons (a, b) e -> c -> T -> d) ==> b=d 



-- the actual evaluation
overload eval :: (Lookup (e->x->v))
              => e -> ExpVar x -> v
   where eval e (ExpVar x) = lookup e x
overload eval :: e -> ExpConst n -> n
   where eval e (ExpConst n) = n
overload eval :: (Eval (e->a1->v1), Eval (e->a2->v2))
              => e -> ExpPair a1 a2 -> (v1,v2)
   where eval e (ExpPair a1 a2) = (eval e a1, eval e a2)


overload eval :: (Eval (e->a->(v1,v2)))
              => e -> ExpPi1 a -> v1
   where eval e (ExpPi1 a) = fst ((eval e a)::(v1,v2))


overload eval :: (Eval (e->a->(v1,v2)))
              => e -> ExpPi2 a -> v2
   where eval e (ExpPi2 a) = snd ((eval e a)::(v1,v2))



overload eval :: (Eval (Cons (x,v) e -> a -> v'))
              => e -> ExpAbs x a -> (v->v')
   where eval e (ExpAbs x a) = \v -> eval (Cons (x,v) e) a


overload eval :: (Eval (e->a1->(v1->v2)), Eval (e->a2->v1))
              => e -> ExpApp a1 a2 -> v2
   where eval e (ExpApp a1 a2) = (eval e a1) ((eval e a2)::v1)

--rule Eval (e->ExpApp a1 a2->v2) ==> Eval (e->D a1->(v1->v2)), Eval (e->D a2->v1)


{- simply-typed version

overload eval ::  (Eval (e->e1->v1), Eval ((Cons (x,v1) e)->e2->v2))
          => e -> ExpLet x e1 e2 -> v2
         eval e (ExpLet x e1 e2) = eval (Cons (x,(eval e e1::v1)) e) e2
-}

{- this approach would work as well, however, we need multi-headed simps

overload tok

rule Eval (e->(ExpPLet x e1 e2)->v) <==> Tok(e,x,e1), Eval(e->e2->v)
rule Tok(e,x,e1), Eval(e'->(ExpLVar x)->t1) <==> Tok(e,x,e1), Eval(e'->e1->t1)
-}

--rule Eval (e->(ExpPLet x e1 e2)->v) <==> Eval((Cons (x,(e,e1)) e)->e2->v)
overload eval :: (Eval((Cons (x,(e,e1)) e)->e2->v)) => e->(ExpPLet x e1 e2)->v
   where eval e (ExpPLet x e1 e2) = eval ((Cons (x,(e,e1))) e) e2


--rule Eval (e'->(ExpLVar x)->t1) <==> Lookup (e'->x->(e,e1)), Eval (e->e1->t1)


overload eval :: (Lookup (e->x->(e',e1)), Eval (e'->e1->t1)) 
               => e -> (ExpLVar x) -> t1
   where eval e (ExpLVar x) = eval (fst ((lookup e x)::(e',e1))) 
                                   (snd ((lookup e x)::(e',e1)))


{- note that we don't allow for
   local let definitions at the moment

overload eval :: (Lookup (e->x->(e',e1)), Eval (e'->e1->t1)) 
               => e -> (ExpLVar x) -> t1
         eval e (ExpLVar x) = let ((e',e1)::(e',e1))= lookup e x
                                in eval e' e1

-}



--rule Eval (e->(ExpVar x)->v)      ==> Lookup (e->x->v)
rule Eval (e->(ExpConst n)->v)    ==> v =n
rule Eval (e->(ExpPair a1 a2)->v) ==> v = (v1,v2)
--                                      Eval (e->a1->v1), Eval (e->a2->v2)
--rule Eval (e->(ExpPi1 a)->v1)     ==> Eval (e->a->(v1,v2))
--rule Eval (e->(ExpPi1 a)->v2)     ==> Eval (e->a->(v1,v2))
rule Eval (e->(ExpAbs x a)->v)    ==> v = v1->v2
--                                    Eval ((Cons (x,v1) e)->a->v2)
---rule Eval (e->(ExpApp a1 a2)->v1) ==> Eval (e->a1->(v2->v1)), Eval (e->a2->v2)
-- commented out rules yield non-confluent set of CHRs



-- test:
-- exp :: ExpApp (ExpConst Int) (ExpConst Int)
exp =  ExpApp (ExpConst (S Z)) (ExpConst (S (S Z)))
--
--res = eval Nil exp


x1 = Zero
x2 = Succ x1
x3 = Succ x2
one = S Z



exp2 = ExpApp (ExpAbs x1 (ExpVar x1)) (ExpConst one)
res2 = eval Nil exp2


exp3 = ExpPLet x2 (ExpAbs x1 (ExpVar x1)) 
        (ExpApp (ExpLVar x2) (ExpLVar x2))
--      (ExpApp (ExpVar x2) (ExpAbs x3 (ExpVar x3))) 
res3 = eval Nil exp3




exp4 = ExpApp (ExpAbs x1 (ExpVar x1)) (ExpAbs x1 (ExpVar x1))
res4 = eval Nil exp4

exp5 = ExpAbs x1 (ExpVar x1)
res5 = eval Nil exp5


{-

rule Eval (e->a->(v1,v2)), Eval (e->a->(v3,v4)) ==> v1=v3, v2=v4

we'd need to prevent that this rule fires, if
     a = ExpVar v
     a = ExpConst ...
this would require disequality guards in Props

-}


{-
-- language of object types
data TypeInt        = TypeInt
data TypePair t1 t2 = TypePair t1 t2
data TypeFunc t1 t2 = TypeFunc t1 t2

overload reify :: TypeInt
         reify =  TypeInt
overload reify :: (Reify t1, Reify t2) => TypePair t1 t2
         reify =  TypePair reify reify
overload reify :: (Reify t1, Reify t2) => TypeFunc t1 t2
         reify =  TypeFunc reify reify


-- type inference
overload infer :: (Lookup (e->x->t), Reify t)
               => e -> ExpVar x -> t
         infer e (ExpVar x) = reify 
overload infer :: e -> ExpConst Int -> TypeInt
         infer e (ExpConst n) = (reify :: TypeInt)
overload infer :: (Infer (e->a1->t1), Infer (e->a2->t2), Reify t1, Reify t2)
               => e -> ExpPair a1 a2 -> TypePair t1 t2
         infer e (ExpPair a1 a2) = reify
overload infer :: (Infer (e->a->TypePair t1 t2), Reify t1)
               => e -> ExpPi1 a -> t1
         infer e (ExpPi1 a) = reify
overload infer :: (Infer (e->a->TypePair t1 t2), Reify t2)
               => e -> ExpPi2 a -> t2
         infer e (ExpPi2 a) = reify
overload infer :: (Infer (Cons (x,t1) e->a->t2), Reify t1, Reify t2)
               => e -> ExpAbs x a -> TypeFunc t1 t2
         infer e (ExpAbs x a) = reify
overload infer :: (Infer (e->a1->TypeFunc t2 t), Infer (e->a2->t2), Reify t)
               => e -> ExpApp a1 a2 -> t
         infer e (ExpApp a1 a2) = reify




--rule Infer (e->(ExpVar x)->t)      ==> Lookup (e->x->t)
rule Infer (e->(ExpConst n)->t)    ==> t = TypeInt
--rule Infer (e->(ExpPair a1 a2)->t) ==> t = TypePair t1 t2
--                                       Infer (e->a1->t1), Infer (e->a2->t2)
--rule Infer (e->(ExpPi1 a)->t1)     ==> Infer (e->a->(TypePair t1 t2))
--rule Infer (e->(ExpPi2 a)->t2)     ==> Infer (e->a->(TypePair t1 t2))
rule Infer (e->(ExpAbs x a)->t)      ==> t = TypeFunc t1 t2
--                                       Infer ((Cons (x,t1) e)->a->t2)
--rule Infer (e->(ExpApp a1 a2)->t)  ==> Infer (e->a1->(TypeFunc t2 t)),
--                                       Infer (e->a2->t2)



-- combined inference and evaluation
overload
exec :: (Eval (env->exp->val), Infer (tenv->exp->typ))
     => tenv -> env -> exp -> (typ, val)
exec tenv env exp = (infer tenv exp, eval env exp)

rule Exec t ==> t=t1->t2->e->(a,b) 

ex2 = exec Nil Nil exp2 

ex5 = exec Nil Nil exp5
-}
