{-# OPTIONS -fglasgow-exts -fallow-undecidable-instances #-}

import Prelude hiding (lookup)

main = undefined

-- language of object expressions
data ExpVar x      = ExpVar x
data ExpConst n    = ExpConst n
data ExpPair a1 a2 = ExpPair a1 a2
data ExpPi1 a      = ExpPi1 a
data ExpPi2 a      = ExpPi2 a
data ExpAbs x a    = ExpAbs x a
data ExpApp a1 a2  = ExpApp a1 a2

-- result of type-level computations
data T = T;
data F = F;

-- we use numbers to model variables
data Zero   = Zero;
data Succ n = Succ n;

class Equal a b c | a b -> c where eq :: a -> b -> c
instance Equal Zero (Succ a) F where
         eq Zero (Succ a) = F;
instance Equal (Succ a) Zero F where
         eq (Succ a) Zero = F;
instance Equal Zero Zero T where
         eq Zero Zero = T;
instance (Equal a b c) => Equal (Succ a) (Succ b) c where
         eq (Succ a) (Succ b) = eq a b;

-- we use lists for the environment
data Nil      = Nil;
data Cons a b = Cons a b;

class Lookup e x y | e x -> y where lookup :: e -> x -> y
instance (Lookup' (Cons (x1,v) e) x2 b v', Equal x1 x2 b) =>
          Lookup (Cons (x1,v) e) x2 v' where
         lookup (Cons (x1,v) e) x2 = lookup' (Cons (x1,v) e) x2 (eq x1 x2 :: b);

class Lookup' e x b y | e x b -> y where lookup' :: e -> x -> b -> y
instance Lookup' (Cons (x1,v) e) x2 T v where
         lookup' (Cons (_,v) e) _ T = v;
instance (Lookup e x2 v') => Lookup' (Cons (x1,v) e) x2 F v' where
         lookup' (Cons (x1,v) e) x2 F = lookup e x2;


-- the actual evaluation
class Eval e a v {-| e a -> v-} where eval :: e -> a -> v
instance (Lookup e x v) => Eval e (ExpVar x) v where
         eval e (ExpVar x) = lookup e x;
instance Eval e (ExpConst n) n where
         eval e (ExpConst n) = n;
instance (Eval e a1 v1, Eval e a2 v2) => Eval e (ExpPair a1 a2) (v1,v2) where
         eval e (ExpPair a1 a2) = (eval e a1, eval e a2);
instance (Eval e a (v1,v2)) => Eval e (ExpPi1 a) v1 where
         eval e (ExpPi1 a) = fst (eval e a :: (v1,v2));
instance (Eval e a (v1,v2)) => Eval e (ExpPi2 a) v2 where
         eval e (ExpPi2 a) = snd (eval e a :: (v1,v2));
instance (Eval (Cons (x,v) e) a v') => Eval e (ExpAbs x a) (v->v') where
         eval e (ExpAbs x a) = \v -> eval (Cons (x,v) e) a;
instance (Eval e a1 (v2->v), Eval e a2 v2) => Eval e (ExpApp a1 a2) v where
         eval e (ExpApp a1 a2) = (eval e a1) (eval e a2 :: v2);


-- test:
-- fine  = eval Nil (ExpApp (ExpAbs 1 (ExpVar 1)) (ExpConst 2))
-- bogus = eval Nil (ExpApp (ExpConst 1) (ExpConst 2))


-- language of object types
data TypeInt        = TypeInt;
data TypePair t1 t2 = TypePair t1 t2;
data TypeFunc t1 t2 = TypeFunc t1 t2;

class Reify a where reify :: a
instance Reify TypeInt where
         reify = TypeInt;
instance (Reify t1, Reify t2) => Reify (TypePair t1 t2) where
         reify = TypePair reify reify;
instance (Reify t1, Reify t2) => Reify (TypeFunc t1 t2) where
         reify = TypeFunc reify reify;


-- type inference
class Reify t => Infer e a t {-| e a -> t-} where infer :: e -> a -> t
instance (Lookup e x t, Reify t) => Infer e (ExpVar x) t where
         infer e (ExpVar x) = reify
instance Infer e (ExpConst Int) TypeInt where
         infer e (ExpConst n) = reify
instance (Infer e a1 t1, Infer e a2 t2) =>
          Infer e (ExpPair a1 a2) (TypePair t1 t2) where
         infer e (ExpPair a1 a2) = reify
instance (Infer e a (TypePair t1 t2), Reify t1) => Infer e (ExpPi1 a) t1 where
         infer e (ExpPi1 a) = reify
instance (Infer e a (TypePair t1 t2), Reify t2) => Infer e (ExpPi2 a) t2 where
         infer e (ExpPi2 a) = reify
instance (Infer (Cons (x,t1) e) a t2, Reify t1) =>
          Infer e (ExpAbs x a) (TypeFunc t1 t2) where
         infer e (ExpAbs x a) = reify
instance (Infer e a1 (TypeFunc t2 t), Infer e a2 t2, Reify t) =>
          Infer e (ExpApp a1 a2) t where
         infer e (ExpApp a1 a2) = reify
{-
instance (Lookup e x t) => Infer e (ExpVar x) t where
         infer e (ExpVar x) = lookup e x;
instance Infer e (ExpConst Int) TypeInt where
         infer e (ExpConst n) = TypeInt;
instance (Infer e a1 t1, Infer e a2 t2) =>
          Infer e (ExpPair a1 a2) (TypePair t1 t2) where
         infer e (ExpPair a1 a2) = TypePair (infer e a1) (infer e a2);
instance (Infer e a (TypePair t1 t2)) => Infer e (ExpPi1 a) t1 where
         infer e (ExpPi1 a) = t1 where TypePair t1 (t2::t2) = infer e a;
instance (Infer e a (TypePair t1 t2)) => Infer e (ExpPi2 a) t2 where
         infer e (ExpPi2 a) = t2 where TypePair (t1::t1) t2 = infer e a;
instance (Infer (Cons (x,t1) e) a t2, Reify t1) =>
          Infer e (ExpAbs x a) (TypeFunc t1 t2) where
         infer e (ExpAbs x a) = TypeFunc t1 (infer (Cons (x,t1) e) a)
				where t1 = reify
instance (Infer e a1 (TypeFunc t2 t), Infer e a2 t2) =>
          Infer e (ExpApp a1 a2) t where
         infer e (ExpApp a1 a2) = t where TypeFunc (t2::t2) t = infer e a1
-}

-- combined inference and evaluation
exec :: (Eval env exp val,Infer tenv exp typ) => tenv -> env -> exp -> (typ,val)
exec tenv env exp = (infer tenv exp, eval env exp);
