

-- Modelling kind checking in Chamleon

-- Predicate kind
overload kind

data Star = Star

-- We introduce internal encodings of
data L a = L a   -- list,
data A a = A a   -- arrow, and
data P a = P a   -- pair constructor

-- we need them, because in our current syntax we don't allow for partial application of
-- arrow and pair constructors. Therefore, we will transform types from external to
-- internal syntax. See below.

-- Kinding rules

rule Kind (A,k) ==> k=Star->Star->Star
rule Kind (P,k) ==> k=Star->Star->Star
rule Kind (L,k) ==> k=Star->Star

-- We need to provide kinding rules for each user-defined constructor
data Int = One
rule Kind (Int,k) ==> k=Star

-- Booleans are builtin
rule Kind (Bool,k) ==> k=Star

data T a b = Blah
rule Kind (T,k) ==> k=Star->Star->Star

-- Handling constructor application
rule Kind (c c',k'') ==> Kind (c,k), Kind (c',k'), k=k'->k''


-- Transforming external into internal syntax for kind checking

overload trans

rule Trans (t->t') <==> Kind ((A t) t',k)
rule Trans ((t,t')) <==> Kind ((P t) t',k)
rule Trans ([t]) <==> Kind (L t,k)

-- For each user-defined constructor we need to add two transformation rules
rule Trans (T) <==> Kind (T, k)
rule Trans (T t) <==> Kind (T t,k)

rule Trans (Int) <==> Kind (Int,k)
rule Trans (Int t) <==> Kind (Int t,k)


-- Kind checking

{-
overload check

rule Check t ==> Trans (T a -> (T a a))
--rule Check t ==> Trans (Int->Int)
--rule Check t ==> Trans ((a c)-> (a b))

exp = check
-}


overload annot1
rule Annot1 t <==> t = T a a
v1 = annot1

overload annot2
rule Annot2 t <==> t = T a b
v2= annot2


overload check
--rule Check t ==> t=a->a, Trans (a)

--exp = check (v1,v2)


rule Check t ==> t=(b->c)->(f b->f c), Trans t
exp = check

{- In general, any expression

   f = e 

 will be translated to

  f = check e

where

overload check

rule Check t ==> t=a->a, Kind (a,k)

-}


