
-- We introduce a new overloaded identifier and declare that this identifier
-- must always be of a particular shape. Corresponds to
-- class Bind2 ma a mb where bind2 :: ma->(a->mb)->mb 
-- in Haskell

overload bind2
rule Bind2 x  ==> x=ma->(a->mb)->mb

-- The following CHR enforces the functional dependency

rule Bind2 (ma->(a->mb)->mb), Bind2 (ma->(b->mc)->mc) ==> a=b, mb=mc

-- Side issue. Above CHR might too strong, e.g.
-- x1 = do read ; write ; ...
-- x2 = do read ; read ;
-- where
-- read :: M Read ()
-- write :: M Write ()
-- Seems we need to overloda bind here, e.g.
-- bind :: M Read -> (() -> M Write) -> M (Read,Write)
-- bind :: M Read -> (() -> M Read) -> M (Read,Read)
-- Suggest the following generalization
-- bind :: ma -> (a -> mb) -> mc
-- where ma, mb and mc are monads

-- An example instance

data E a = Succ a | Err


overload bind2 :: E a -> (a -> E b) -> E b where
         bind2 (Succ a) k = k a
         bind2 Err k      = Err
-- In addition, we'd like to state that this instance must satisfy
-- (Bind2 (mb->(b->mc)->mc)) => ma->(a->mb)->mb

-- In Chameleon, the FD must be enforced for each particular instance.
-- That's what the rule below is doing.

rule Bind2 (E a -> x) ==> x = (a -> E b) -> E b

-- We yet need to check that every instance ma->(a->mb)->mb comes with a 
-- Bind2 (mb->(b->mc)->mc. That is, the type sigma of an instance
-- must be a subtype of Bind2 (mb->(b->mc)->mc) => ma->(a->mb)->mb
-- We could use newtype and first-order types, but we can do this without
-- such extensions.
-- Assume sigma is the declared type and sigma' is the type which needs to be satisfied
-- The following piece code will do the checking for us

-- We'd also need to check that mb and ma, share the same type constructor

isSigma :: (Bind2 (mb->(b->mc)->mc)) => ma->(a->mb)->mb
isSigma = undefined

checkSigma' :: E a -> (a -> E b) -> E b
checkSigma' = isSigma

rule Bind2 (E a->(a->mb)->mb), Bind2 (E a->(b->mc)->mc) ==> a=b, mb=mc

-- THAT'S THE PROPER WAY TO CHECK
hasType :: (Bind2 (E b->(b->mc)->mc)) => E a->(a->E b)->E b
hasType = isType

-- works because in -d mode we employ the more general subsumption test !!!!!!!
isType :: E a -> (a -> E b) -> E b
isType = undefined

-- we'd like to have a slightly less restrictive subsumption test (compared to the one
-- used for the evidence translation)
-- exists^1 variables are allowed to instantiated and certain constraints to be removed!!

overload return

data Maybe a = Just a | Nothing

{-
rule Bind2 (Maybe a ->(b->mc)->mc) ==> b=a, mc=E c

overload bind2 :: (Bind2 (mb->(a->mb)->mb), Return (a->mb)) => Maybe a ->(a->mb)->mb where
         bind2 (Just a) k = k a
         bind2 Nothing k      = bind2 ((return (undefined::a))::mb) k
                               -- scoped type variable necessary, otherwise we infer an
                               -- ambiguous type scheme

rule Return (b->mb), Return (c->mb) ==> b=c

-}

-- checking



{-  OLD STUFF from previous attempts

-- We define the signature which must be subsumed
check :: (Bind2 (mb->(b->mc)->mc)) => ma->(a->mb)->mb;
check = undefined;

-- Some auxiallary definitions.
forget :: a->b->a;
forget x y = x;

overload mycheck;
rule Mycheck x <==> x=a->a->a;

-- The bind2 instances from above
g :: E a -> (a-> E b)-> E b;
g (Succ a) k = k a;
g Err k      = Err;

-- we need the type signature for both types
-- to check for entailment
--erk :: E a -> (a-> E b)-> E b;
--erk = forget g (mycheck check g);
--erk = subsumes g check;

subsumes g f = forget g (mycheck f g);

subsumes2 g f = let t = mycheck f g;
                in g;
id x = x;

data Int = Int;

test :: Int->Int;
test = undefined;

-}