

-- Modeling extensible records

-- we use singleton list to model records

data Nil = Nil
data Cons x xs = Cons x xs

-- we use singleton numbers to model record labels

data Zero   = Zero
data Succ n = Succ n

-- result of type level computations
data T = T
data F = F

overload eq :: Zero -> Zero -> T where
         eq Zero Zero = T
overload eq :: (Eq (a->b->c)) => Succ a -> Succ b -> c where
         eq (Succ a) (Succ b) = eq a b
overload eq :: Zero -> Succ a -> F where
         eq Zero (Succ a) = F
overload eq :: Succ a -> Zero -> F where
         eq (Succ a) Zero = F

-- improvement
rule Eq (a->b->c), Eq (a->b->d) ==> c=d
rule Eq (Zero->Zero->a)         ==> a=T
rule Eq (Succ a->Zero->b)       ==> b=F 
rule Eq (Zero->Succ a->b)       ==> b=F     

-- record selection

overload select :: (Select' (Cons (x1,v) e -> x2 -> b -> v'), Eq (x1->x2->b))
                => Cons (x1,v) e -> x2 -> v' where
         select (Cons (x1,v) e) x2 = select' (Cons (x1,v) e) x2 (eq x1 x2)
overload select' :: Cons (x1,v) e -> x2 -> T -> v where
         select' (Cons (_,v) e) _ T = v
overload select' :: (Select (e->x2->v')) => Cons (x1,v) e -> x2 -> F -> v' where
         select' (Cons (x1,v) e) x2 F = select e x2

rule Select' (Cons (a, b) e -> c -> T -> d) ==> b=d 


-- record extension, we don't check whether a field already exists

overload extend :: (l,x)->Nil->(Cons (l,x) Nil) where
         extend (l,x) Nil = Cons (l,x) Nil

overload extend :: (l,x) -> (Cons (l',x') xs) -> Cons (l,x) (Cons (l',x') xs) where
         extend (l,x) (Cons (l',x') xs) = Cons (l,x) (Cons (l',x') xs)

rule Extend a ==> a=(l,x)->r->r'
rule Extend (a->b->r), Extend(a->b->r') ==> r=r'
rule Extend((l,x)->Nil->t) ==> t=Cons (l,x) Nil
rule Extend((l,x)->(Cons (l',x') xs->t)) ==> t=Cons (l,x) (Cons (l',x') xs)

-- check that all record labels are unique

overload check :: (Unique x) => x->x where
         check x = x

rule Check a ==> a=t->t
 
overload unique
overload unique'
overload unique''

rule Unique (Cons (l,x) (Cons (l',x') xs)) ==> Unique'(l,(Cons (l',xs') xs)), Unique'(l',xs)
rule Unique'(l,Cons (l',x) xs)             ==> Eq(l->l'->a), Unique''(a->l->xs)
rule Unique''(F->l->xs)                    ==> Unique'(l,xs)
rule Unique''(T->l->xs)                    ==> T=F

-- field removal, we don't check whether a field doesn't exist

overload remove :: l -> Nil -> Nil where
         remove l Nil = Nil

overload remove :: (Eq (l->l'->a), Remove' (a->l->(Cons (l',v) r)->r')) 
                 => l -> (Cons (l',v) r) -> r' where
         remove l (Cons (l',v) r) = remove' (eq l l') l (Cons (l',v) r)

rule Remove a ==> a=l->r->r'
rule Remove (l->r->r'), Remove (l->r->r'') ==> r'=r''
rule Remove (l->Nil->r') ==> r'=Nil

overload remove' :: T -> l -> (Cons (l',v) r) -> r where
         remove' T l (Cons (l',v) r) = r

overload remove' :: (Remove (l->r->r')) => F -> l -> (Cons (l',v) r) -> (Cons (l',v) r') where
         remove' F l (Cons (l',v) r) = Cons (l',v) (remove l r)

-- improvement
rule Remove' (a->c->(Cons (l',v) r''')->r'), Remove' (a->c->(Cons (l',v) r''')->r'') ==> r'=r''
rule Remove' (T->a->(Cons (b,c) d)->e) ==> e=d
rule Remove' (F->a->(Cons (b,c) d)->e) ==> e=Cons (b,c) e'



-- examples

l0 = Zero
l1 = Succ l0
l2 = Succ l1

data String = A | B | C
data Phone  = P1 | P2

r1 = Nil
r2 = extend (l0,A) r1

r2' = check r2
r3 = extend (l1,P1) r2

x = select r2 l0

y = remove l0 r3

z = extend (l1, P2) r3
-- type (check z)
-- yields an error